****************************************************
* This class provides a set of tools for exploring
* and cleaning up a project. To use it, instantiate
* the class to a variable. 
*
* Step 1 is always calling the GetProject method, passing
* the name of the project you want to work on.
* 
* A number of methods (all with names beginning with "List")
* simply survey the project for information,
* putting the results into a cursor. In each case, you can pass
* the alias you want to use as a parameter; the method returns
* the alias it actually uses. The List methods are:
*
* ListFilesInRoot - fills a cursor with a list of files in the project
*                   that reside in the project's root directory.
* ListFolders - fills a cursor with a list of the folders
*               that contain the files in the project.
* ListGraphics - fills a cursor with a list of the graphic files
*                referenced by Picture and Icon properties of forms
*                and classes in the project.
* ListMissingFiles - fills a cursor with a list of files named
*                    in the project that are not found on disk.
* ListOutsideFiles - fills a cursor with a list of files included
*                    in the project that are located in folders 
*                    outside the project tree (that is, not below
*                    the project's home directory).
* ListUnusedFiles - fills a cursor with a list of files in the 
*                   project's tree that are not included in the project.
*                   Accepts an option second parameter of a comma-separated 
*                   list of folders to exclude from the search.
*
* Each of the List methods has a corresponding Report method
* that takes its results and displays them in a report. Each of the
* Report methods accepts two parameters (except as noted): the alias to 
* report on and, optionally, the name of a report file to run. If two
* second parameter is omitted, a quick report is created, used, and deleted.
* The Report methods are:
*
* ReportFilesInRoot 
* ReportFolders 
* ReportGraphics
* ReportMissingFiles 
* ReportOutsideFiles 
* ReportUnusedFiles - This method accepts an optional second parameter
*                     (before the optional report parameter) to indicate
*                     the sort order. Pass 1 to sort by path and filename;
*                     pass 2 to sort by filename, then path.
*
* A few methods make changes to the project. They are:
* 
* CopyProject - accepts a path as parameter and makes a complete copy of 
*               the project and its files in the specified location, maintaining
*               the folder structure of the original project.
* MoveFilesFromRoot - moves project files that are in the project's home directory
*                     into the appropriate child folders, based on the file's
*                     extension. Prompts for each folder the first time it needs it.
*                     Relative paths within forms and classes are updated.
*                     WARNING: relative path updates are performed only within a single
*                        file, so if one moved file refers to another moved file,
*                        the references to the second in the first will not be fixed.
*                        In addition, this method moves Include files, but doesn't 
*                        update the files that refer to them.
*
* The class also includes a number of worker methods, used by other methods:
*
* BuildFileCursor - fills a cursor with the list of files in the project.
* Destroy - closes the referenced project.
* FixClassLoc - fixes relative references in the ClassLoc field of forms and classlibs.
* FixGraphics - fixes relative references in the Picture and Icon properties of forms 
*               and classlibs.
* GetNewPath - build a new relative path to a file based on old and new root directories
* GetValidAlias - accepts a suggested alias and a default alias, and returns a string that
*                 is a valid alias. If the suggested alias is valid, it's returned. If not,
*                 and the default alias is valid, it's returned. Otherwise, a random
*                 alias is created and returned.
* ListUnusedFilesInOneFolder - recursive method called by ListUnusedFiles, it checks
*                              all files in the specified folder, and calls itself
*                              to handle subfolders.
* ReportOnProject - accepts an alias, a list of files to report on, and optionally, the 
*                   name of a report file, and produces a report on the specified alias.
*                   If the report file parameter is omitted or the report doesn't exist,
*                   creates a quick report that includes the specified fields. (If the 
*                   list of fields is omitted, all fields are included.)
* ValidAlias - checks a string to see whether it's a valid alias.
*
* Author: Tamar E. Granor
* Written: March, 2011
****************************************************

DEFINE CLASS ProjectTools AS Custom

oProject = .null.
cHomeDir = ""
cFormDir = ""
cClassLibDir = ""
cProgDir = ""
cMenuDir = ""
cGraphicsDir = ""
cReportDir = ""
cIncludeDir = ""

****************************************************
PROCEDURE GetProject(cProject)
* Connect a project to this object.

LOCAL lProceed

* If we already have a project, close it
IF NOT ISNULL(This.oProject)
	This.oProject.Close()
ENDIF 


IF VARTYPE(m.cProject) = "C" 
	* Open the project
	TRY 
		MODIFY PROJECT (m.cProject) NOWAIT 
		This.oProject = _VFP.ActiveProject
		lProceed = .t.
	CATCH 
		lProceed = .F.
	ENDTRY
ELSE
	lProceed = .F.
ENDIF 

IF NOT m.lProceed
	MESSAGEBOX("Unable to open specified project: " + m.cProject + ;
		           ". Make sure to provide a good path.")
	RETURN
ENDIF

This.cHomeDir = This.oProject.HomeDir

* Clear all the folder references
This.cFormDir = ""
This.cClassLibDir = ""
This.cProgDir = ""
This.cMenuDir = ""
This.cGraphicsDir = ""
This.cReportDir = ""
This.cIncludeDir = ""

RETURN
ENDPROC

****************************************************
PROCEDURE Destroy

* Close the project if we have one
IF NOT ISNULL(This.oProject)
	This.oProject.Close()
ENDIF 

RETURN
ENDPROC


****************************************************
PROCEDURE BuildFileCursor(cAlias, lIndex)
* Build a cursor of files in the project.
* If indicated, index it on the path + file combination.

cAlias = This.GetValidAlias(m.cAlias, "csrProjectFiles")

CREATE CURSOR (m.cAlias) (iID I AUTOINC, mFile M, mPath M)
INDEX on iID TAG iID
IF m.lIndex
	INDEX on UPPER(CAST(FORCEPATH(mFile, mPath) AS C(200))) TAG FilePath
ENDIF 

LOCAL oFile, cFilename, cPath

FOR EACH oFile IN This.oProject.Files 
	cFileName = JUSTFNAME(oFile.Name)
	cPath = JUSTPATH(oFile.Name)
	INSERT INTO (m.cAlias) ;
		(mFile, mPath) ;
		VALUES (m.cFileName, m.cPath)
ENDFOR  

RETURN 
ENDPROC

****************************************************
PROCEDURE ListUnusedFiles(cAlias, cExcludeFolders)
* Identify files in project folders that are not
* in the project. Don't look in folders listed
* in cExcludeFolders.

IF VARTYPE(m.cExcludeFolders) <> "C"
	cExcludeFolders = ""
ELSE
	* Bracket with commas, so we can search for the 
	* specified folder with commas on both ends.
	* Change to upper to simplify searching.
	* Also, prefix each with the project's
	* home directory. 
	cExcludeFolders = "," + This.cHomeDir + "\" + ;
	                  UPPER(ALLTRIM(STRTRAN(m.cExcludeFolders, ",", "," + This.cHomeDir + "\"))) ;
	                  + ","
ENDIF 

cAlias = This.GetValidAlias(m.cAlias, "csrUnusedFiles")
   	
CREATE CURSOR (m.cAlias) ;
	(iID I AUTOINC, mFileWithPath M)

* First, build a cursor of files in the project.
This.BuildFileCursor("csrProjectFiles", .T.)

This.ListUnusedFilesInOneFolder(This.cHomeDir, m.cAlias, "csrProjectFiles", m.cExcludeFolders)

RETURN m.cAlias
ENDPROC

****************************************************
PROCEDURE ListUnusedFilesInOneFolder(cFolder, cResultAlias, ;
                                     cProjFilesAlias, cExcludeFolders)
* Check files in a specified folder to see whether they're included
* in the project. Drill down recursively, but skip any folder
* listed in cExcludeFolders.

LOCAL aFolderFiles[1], nFileCount, nFile, cFileName, cSubFolder

nFileCount = ADIR(aFolderFiles, FORCEPATH("*.*", m.cFolder), "D") 

FOR nFile = 1 TO m.nFileCount
	cFileName = aFolderFiles[m.nFile, 1]
	IF NOT INLIST(cFileName, ".", "..")
		IF "D" $ aFolderFiles[m.nFile, 5] 
			* It's a folder, so call this method recursively
			* unless it's excluded
			cSubfolder = UPPER(ALLTRIM(FORCEPATH(m.cFileName, m.cFolder)))
			IF NOT "," + m.cSubFolder + "," $ m.cExcludeFolders
				This.ListUnusedFilesInOneFolder(m.cSubFolder, m.cResultAlias, m.cProjFilesAlias, m.cExcludeFolders)
			ENDIF 
		ELSE
			* Check for this file. 
			cExt = JUSTEXT(m.cFileName)
			IF INLIST(m.cExt, "SCT", "VCT", "MNT", "FRT", "LBT")
				cAssocExt = LEFT(m.cExt, 2) + "X"
				cFileToSearch = FORCEEXT(JUSTSTEM(m.cFileName), m.cAssocExt)
			ELSE
				cFileToSearch = m.cFileName
			ENDIF
			
			cFileToSearch = FORCEPATH(m.cFileToSearch, m.cFolder)
			IF NOT SEEK(m.cFileToSearch, m.cProjFilesAlias, "FilePath")
				INSERT INTO (m.cResultAlias) ;
					(mFileWithPath) ;
					VALUES ;
					(FORCEPATH(m.cFileName, m.cFolder))
			ENDIF
			
		ENDIF
	ENDIF 	
ENDFOR 

RETURN RECCOUNT(m.cResultAlias)
ENDPROC

****************************************************
PROCEDURE ReportUnusedFiles(cAlias, nOrder, cReport)
* Produce a report of unused files in the project folders.
* cAlias is the cursor containing the list.
* nOrder indicates the order for the report. It it's omitted or 1,
* the report is sorted by the fully pathed filename.
* If it's 2, the report is sorted by filename, then by path.

LOCAL cOrder

DO CASE 
CASE VARTYPE(m.nOrder) <> "N" OR m.nOrder = 1
	cOrder = "cFilePath, cFileName"

CASE m.nOrder = 2
	cOrder = "cFileName, cFilePath"
	
OTHERWISE 
	cOrder = "cFilePath, cFileName"
ENDCASE 
	
* Select the data into a new cursor, sorting as requested.
SELECT mFileWithPath AS UnusedFile, ;
       CAST(JUSTFNAME(mFileWithPath) as C(250)) as cFileName, ;
       CAST(JUSTPATH(mFilewithPath) as c(250)) as cFilePath ;
	FROM (m.cAlias) ;
	ORDER BY &cOrder ;
	INTO CURSOR csrReportUnused

This.ReportOnProject("csrReportUnused", "UnusedFile", m.cReport)	

USE IN SELECT("csrReportUnused")

RETURN
ENDPROC

****************************************************
PROCEDURE ListFilesInRoot(cAlias)
* Make a list of files in the project that are in the 
* home directory of the project.

LOCAL oFile

cAlias = This.GetValidAlias(m.cAlias, "csrFilesInRoot")

CREATE CURSOR (m.cAlias) ;
	(iID I AUTOINC, mFileName M)
	
LOCAL cUpperHome, cFileName

cUpperHome = UPPER(This.cHomeDir)

FOR EACH oFile IN This.oProject.Files 
	cFileName = UPPER(oFile.Name)
	IF JUSTPATH(m.cFileName) == m.cUpperHome
		INSERT INTO (m.cAlias) ;
			(mFileName) ;
			VALUES ;
			(oFile.Name)
	ENDIF 
ENDFOR

RETURN m.cAlias
ENDPROC

****************************************************
PROCEDURE ReportFilesInRoot(cAlias, cReport)
* Report on the list of files in the root directory.

SELECT mFileName, LEFT(mFileName, 250) as cFileName ;
	FROM (m.cAlias) ;
	ORDER BY 2 ;
	INTO CURSOR csrReport
	
This.ReportOnProject("csrReport", "mFileName", m.cReport)

USE IN SELECT("csrReport")

RETURN
ENDPROC

****************************************************
PROCEDURE MoveFilesFromRoot
* Move files that reside in the root folder of the 
* project into the appropriate folder. The folders
* are specified by project properties. If those
* folder properties aren't filled in, the user
* will be prompted the first time any is needed.
* Returns the number of files skipped.

LOCAL nFilesToMove, nFile, cFile, cPartnerFile
LOCAL cDestFolder, oFile, cType, nSkipped, cDescription

nFilesToMove = This.ListFilesInRoot("csrFilesInRoot")

* Add a field to track whether we've successfully moved
* the file.
ALTER table csrFilesInRoot ADD lMoved L

nSkipped = 0

SELECT csrFilesInRoot
SCAN 
	cFile = csrFilesInRoot.mFileName
	IF FILE(m.cFile)		 
		oFile = This.oProject.Files[m.cFile]
	ELSE 
		nSkipped = m.nSkipped + 1 
		LOOP
	ENDIF 
	
	* Strategy is to collect all information,
	* then move the file, remove it from the project,
	* then add it back.
	
	* First, figure out what kind of file we have,
	* thus, where it's going and whether there's a
	* "partner" file.
	
	cType = UPPER(oFile.Type)
	cDestFolder = ""

	DO CASE 
	CASE m.cType = "K"
		IF EMPTY(This.cFormDir)
			This.cFormDir = GETDIR("","Specify form folder for project","Form folder")
		ENDIF 
		
		IF NOT EMPTY(This.cFormDir) 
			cDestFolder = This.cFormDir
			cPartnerFile = FORCEEXT(m.cFile, "SCT")
			
			* Need to fix paths in the file
			This.FixClassLoc(m.cFile, This.cHomeDir, This.cFormDir)
			This.FixPictures(m.cFile, This.cHomeDir, This.cFormDir)

		ENDIF 
		
	CASE m.cType = "V"
		IF EMPTY(This.cClassLibDir)
			This.cClassLibDir = GETDIR("","Specify classlib folder for project","Classlib folder")
		ENDIF 
		
		IF NOT EMPTY(This.cClassLibDir) 
			cDestFolder = This.cClassLibDir
			cPartnerFile = FORCEEXT(m.cFile, "VCT")

			* Need to fix paths in the file
			This.FixClassLoc(m.cFile, This.cHomeDir, This.cFormDir)
			This.FixPictures(m.cFile, This.cHomeDir, This.cFormDir)
		ENDIF 
		
	CASE m.cType = "P"
		IF EMPTY(This.cProgDir)
			This.cProgDir = GETDIR("","Specify program folder for project","Program folder")
		ENDIF 
		
		IF NOT EMPTY(This.cProgDir) 
			cDestFolder = This.cProgDir
			cPartnerFile = ""
		ENDIF 
				
	CASE m.cType = "M"
		IF EMPTY(This.cMenuDir)
			This.cMenuDir = GETDIR("","Specify menu folder for project","Menu folder")
		ENDIF 
		
		IF NOT EMPTY(This.cMenuDir) 
			cDestFolder = This.cMenuDir
			cPartnerFile = FORCEEXT(m.cFile, "MNT")
		ENDIF
				
	CASE m.cType = "R"
		IF EMPTY(This.cReportDir)
			This.cReportDir = GETDIR("","Specify report folder for project","Report folder")
		ENDIF 
		
		IF NOT EMPTY(This.cReportDir) 
			cDestFolder = This.cReportDir
			cPartnerFile = FORCEEXT(m.cFile, "FRT")
		ENDIF 
				
	CASE m.cType = "X" AND ;
	     INLIST(UPPER(JUSTEXT(m.cFile)), ;
	     "ANI", "BMP", "CUR", "DIB", "EMF", "EXIF", "GIF", "GFA", ;
	     "ICO", "JPG", "JPEG", "JPE", "JFIF", "PNG", "TIF", "TIFF", "WMF")
		IF EMPTY(This.cGraphicsDir)
			This.cGraphicsDir = GETDIR("","Specify graphics folder for project","Graphics folder")
		ENDIF 
		
		IF NOT EMPTY(This.cGraphicsDir) 
			cDestFolder = This.cReportDir
			cPartnerFile = ""
		ENDIF 
				
	CASE m.cType = "X" AND ;
	     INLIST(UPPER(JUSTEXT(m.cFile)), "H")
		IF EMPTY(This.cIncludeDir)
			This.cIncludeDir = GETDIR("","Specify include file folder for project","Include file folder")
		ENDIF 
		
		IF NOT EMPTY(This.cIncludeDir) 
			cDestFolder = This.cIncludeDir
			cPartnerFile = ""
		ENDIF 	
	ENDCASE 
	
	IF EMPTY(m.cDestFolder)
		nSkipped = m.nSkipped + 1 
	ELSE 
		cDescription = oFile.Description
		
		cNewFile = FORCEPATH(m.cFile, m.cDestFolder)
		RENAME (m.cFile) TO (m.cNewFile)
		IF NOT EMPTY(m.cPartnerFile) AND FILE(m.cPartnerFile)
			RENAME (m.cPartnerFile) TO FORCEPATH(m.cPartnerFile, m.cDestFolder)
		ENDIF 
		oFile.Remove()
		This.oProject.Files.Add(m.cNewFile)
		
		REPLACE lMoved WITH .T. IN csrFilesInRoot
		
	ENDIF 
		
ENDSCAN 

RETURN "csrFilesInRoot"
ENDPROC 

****************************************************
PROCEDURE FixClassLoc(cFileName, cOldDir, cNewDir)
* In the specified file (either form or classlib), 
* adjust path of ClassLoc field so that it moves
* from relative to cOldDir to relativ to cNewDir

LOCAL cCurPath, cNewPath, cFilePath
LOCAL nOldSelect

nOldSelect = SELECT()

cOldDir = ADDBS(m.cOldDir)
cNewDir = ADDBS(m.cNewDir)

cFilePath = ADDBS(JUSTPATH(m.cFileName))

SELECT 0
USE (m.cFileName) ALIAS __CXFile

SCAN
	IF NOT EMPTY(__CXFile.ClassLoc) 
		cNewPath = This.GetNewPath(__CXFile.ClassLoc, m.cFilePath, m.cNewDir)	
		REPLACE ClassLoc WITH m.cNewPath
	ENDIF
ENDSCAN 

USE IN SELECT("__CXFile")

SELECT (m.nOldSelect)
RETURN
ENDPROC

****************************************************
PROCEDURE FixPictures(cFileName, cOldDir, cNewDir)
* In the specified file (either form or classlib), 
* adjust path of any pictures in the Properties field 
* so that they move from relative to cOldDir to 
* relative to cNewDir

LOCAL cNewPath, cFilePath, cPictPath
LOCAL nOldSelect

nOldSelect = SELECT()

cOldDir = ADDBS(m.cOldDir)
cNewDir = ADDBS(m.cNewDir)

cFilePath = ADDBS(JUSTPATH(m.cFileName))

SELECT 0
USE (m.cFileName) ALIAS __CXFile

SCAN FOR "Picture = " $ __CXFile.Properties OR "Icon = " $ __CXFile.Properties
	IF "Picture = " $ __CXFile.Properties
		* Find the right line and pull the path out
		cPictPath = STREXTRACT(__CXFile.Properties, "Picture = ", CHR(13) + CHR(10))
		cNewPath = This.GetNewPath(m.cPictPath, m.cFilePath, m.cNewDir)
		
		REPLACE Properties WITH STRTRAN(__CXFile.Properties, m.cPictPath, m.cNewPath) IN __CXFile
	ENDIF
	
	IF "Icon = " $ __CXFile.Properties
		* Find the right line and pull the path out
		cPictPath = STREXTRACT(__CXFile.Properties, "Icon = ", CHR(13) + CHR(10))
		cNewPath = This.GetNewPath(m.cPictPath, m.cFilePath, m.cNewDir)
		
		REPLACE Properties WITH STRTRAN(__CXFile.Properties, m.cPictPath, m.cNewPath) IN __CXFile
	ENDIF
ENDSCAN 

USE IN SELECT("__CXFile")

SELECT (m.nOldSelect)
RETURN
ENDPROC

****************************************************
PROCEDURE GetNewPath(cFileWithPath, cOldDir, cNewDir)
* Transform the path for cFileWithPath so that, if it
* was relative to cOldPath, it's now relative to cNewPath

LOCAL cCurPath, cNewPath

* Build the full path to the File
cCurPath = FULLPATH(m.cFileWithPath, m.cOldDir)

* Now find the best relative reference for the new path
cNewPath = SYS(2014, m.cCurPath, m.cNewDir)

RETURN m.cNewPath
ENDPROC

****************************************************
PROCEDURE ListMissingFiles(cAlias)
* Create a list of files in the project that are not
* found in the project folders.

LOCAL oFile, cFilewithPath

cAlias = This.GetValidAlias(m.cAlias, "csrMissingFiles")

CREATE CURSOR (m.cAlias) ;
	(iID I AUTOINC, mFileName M)

FOR EACH oFile IN This.oProject.Files 
	cFileWithPath = oFile.Name
	IF NOT FILE(m.cFilewithPath)
		INSERT INTO (m.cAlias) (mFileName) VALUES (m.cFileWithPath)
	ENDIF
ENDFOR

RETURN m.cAlias

****************************************************
PROCEDURE ReportMissingFiles(cAlias, cReport)
* Produce a report of missing files.

SELECT mFileName, LEFT(mFileName, 250) as cFileName ;
	FROM (m.cAlias) ;
	ORDER BY 2 ;
	INTO CURSOR csrReport
	
This.ReportOnProject("csrReport", "mFileName", m.cReport)

USE IN SELECT("csrReport")

RETURN 

****************************************************
PROCEDURE ListFolders(cAlias)
* Make a list of folders used in the project

cAlias = This.GetValidAlias(m.cAlias, "csrFolders")

CREATE CURSOR (m.cAlias) (iID I AUTOINC, mFolder M, iFileCount I)

LOCAL oFile, cFolder, cUpperFolder

FOR EACH oFile IN This.oProject.Files
	cFolder = JUSTPATH(oFile.Name)
	cUpperFolder = UPPER(m.cFolder)
	SELECT (m.cAlias)
	LOCATE FOR UPPER(ALLTRIM(mFolder)) == m.cUpperFolder
	IF FOUND(m.cAlias)
		REPLACE iFileCount WITH iFileCount + 1 IN (m.cAlias)
	ELSE
		INSERT INTO (m.cAlias) (mFolder, iFileCount) VALUES (m.cFolder, 1)
	ENDIF
ENDFOR 

RETURN m.cAlias

****************************************************
PROCEDURE ReportFolders(cAlias, cReport)
* Produce a report of missing files.

SELECT mFolder, LEFT(mFolder, 250) as cFileName, iFileCount  ;
	FROM (m.cAlias) ;
	ORDER BY 2 ;
	INTO CURSOR csrReport
	
This.ReportOnProject("csrReport", "mFolder,iFileCount", m.cReport)

USE IN SELECT("csrReport")

RETURN 

****************************************************
PROCEDURE ListOutsideFiles(cAlias)
* Make a list of files that are not in 
* this projects folder hierarchy

cAlias = This.GetValidAlias(m.cAlias, "csrOutside")

CREATE CURSOR (m.cAlias) (iID I AUTOINC, mFileName M)

LOCAL oFile

FOR EACH oFile IN This.oProject.Files
	IF NOT (UPPER(This.cHomeDir) $ UPPER(oFile.Name))
		INSERT INTO (m.cAlias) (mFileName) VALUES (oFile.Name)
	ENDIF
ENDFOR 

RETURN m.cAlias
ENDPROC
		
****************************************************
PROCEDURE ReportOutsideFiles(cAlias, cReport)
* Produce a report of files outside the project 
* folder hierarchy.

SELECT mFileName, LEFT(mFileName, 250) as cFileName ;
	FROM (m.cAlias) ;
	ORDER BY 2 ;
	INTO CURSOR csrReport
	
This.ReportOnProject("csrReport", "mFileName", m.cReport)

USE IN SELECT("csrReport")

RETURN 

****************************************************
PROCEDURE ListGraphics(cAlias)
* List all graphics used in project. Include
* whether they're included in project.
* Note that only graphics used in the Picture or Icon properties in 
* forms, classes. 

cAlias = This.GetValidAlias(m.cAlias, "csrGraphics")

CREATE CURSOR (m.cAlias) (iID I AUTOINC, mFileName M, lInProject L, mUsedBy M)

LOCAL oFile, cPictFile, cPictWithPath

FOR EACH oFile IN This.oProject.Files
	IF INLIST(oFile.Type, "K", "V")
		This.ReportToUser("Looking for graphics in " + oFile.Name)
		USE (oFile.Name) ALIAS __CXFile IN 0
	
		SELECT __CXFile
		SCAN FOR "Picture = " $ __CXFile.Properties OR "Icon = " $ __CXFile.Properties
			IF "Picture = " $ __CXFile.Properties 
				cPictFile = STREXTRACT(__CXFile.Properties, "Picture = ", CHR(13) + CHR(10))
				cPictWithPath = FULLPATH(m.cPictFile, oFile.Name)
				
				* Have we listed this file already?
				This.RegisterPicture(m.cPictWithPath, oFile.Name, m.cAlias)
			ENDIF 
			
			IF "Icon = " $ __CXFile.Properties 
				cPictFile = STREXTRACT(__CXFile.Properties, "Icon = ", CHR(13) + CHR(10))
				cPictWithPath = FULLPATH(m.cPictFile, oFile.Name)
				
				* Have we listed this file already?
				This.RegisterPicture(m.cPictWithPath, oFile.Name, m.cAlias)
			ENDIF 
		ENDSCAN 
		
		USE IN SELECT("__CXFile")
	ENDIF
ENDFOR 

RETURN m.cAlias

****************************************************
PROCEDURE RegisterPicture(cPictWithPath, cFile, cAlias)
* Register use of a picture in a specified file.
* cAlias is the alias of the table containing the list.

LOCAL oPictFile, lInProject
LOCAL nOldSelect

nOldSelect = SELECT()
SELECT (m.cAlias)
LOCATE FOR ALLTRIM(mFileName) == m.cPictWithPath
IF FOUND(m.cAlias)
	* Add this file to mUsedBy, unless it's already there\
	IF NOT ((m.cFile + CHR(13) + CHR(10)) $ mUsedBy)
		REPLACE mUsedBy WITH m.cFile + CHR(13) + CHR(10) ADDITIVE IN (m.cAlias)
	ENDIF 
ELSE
	* Is it in the project? 
	TRY 
		oPictFile = This.oProject.Files(m.cPictWithPath)
		lInProject = .T.
	CATCH
		lInProject = .F.
	ENDTRY 
	
	INSERT INTO (m.cAlias) ;
		(mFileName, lInProject, mUsedBy) ;
		VALUES (m.cPictWithPath, m.lInProject, m.cFile + CHR(13) + CHR(10))
ENDIF 

SELECT (m.nOldSelect)

RETURN
ENDPROC

****************************************************
PROCEDURE ReportGraphics(cAlias, cReport)
* Produce a report of graphic files used in the project 

SELECT mFileName, LEFT(mFileName, 250) as cFileName, ;
       lInProject,  mUsedBy ;
	FROM (m.cAlias) ;
	ORDER BY 2 ;
	INTO CURSOR csrReport
	
This.ReportOnProject("csrReport", "mFileName, lInProject, mUsedBy", m.cReport)

USE IN SELECT("csrReport")

RETURN 

****************************************************
PROCEDURE CopyProject(cNewRoot)
* Move a project and all the included files
* from the current folder to the specified
* root, maintaining the current folder structure.

LOCAL oFile, cNewName, cNewPath, cNewProject, oNewProject
LOCAL cMissing, cNoAdd, cFileExt, cNewExt

IF PCOUNT() < 1 OR EMPTY(m.cNewRoot)
	MESSAGEBOX("CopyProject: You must specify the folder for the copy.")
	RETURN
ENDIF 

* First, does the new folder exist
IF NOT DIRECTORY(m.cNewRoot)
	MD (m.cNewRoot)
ENDIF 

* Create the new project.
This.ReportToUser("Creating new project")

cNewProject = FORCEPATH(This.oProject.Name, m.cNewRoot)
CREATE PROJECT (m.cNewProject) NOWAIT 
oNewProject = _VFP.ActiveProject

* Modified 22-April-2011 by TEG
* Need to set HomeDir explicitly
oNewProject.HomeDir = m.cNewRoot

* Modified 13-April-2011 by TEG
* Bring project properties along
WITH oNewProject
	.VersionComments = This.oProject.VersionComments
	.VersionCompany = This.oProject.VersionCompany
	.VersionCopyright = This.oProject.VersionCopyright
	.VersionDescription = This.oProject.VersionDescription
	.VersionLanguage = This.oProject.VersionLanguage
	.VersionNumber = This.oProject.VersionNumber
	.VersionProduct = This.oProject.VersionProduct
	.VersionTrademarks = This.oProject.VersionTrademarks

	.Encrypted = This.oProject.Encrypted
ENDWITH


* Copy all files to appropriate directories
cMissing = ""
cNoAdd = ""
FOR EACH oFile IN This.oProject.Files
	This.ReportToUser("Handling file " + oFile.Name)
	
	IF NOT FILE(oFile.Name)
		* Original file is missing. Make a list for user.
		cMissing = m.cMissing + CHR(13) + CHR(10) + oFile.Name
		LOOP
	ENDIF 
	
	IF This.cHomeDir $ oFile.Name
		cNewName = m.cNewRoot + STREXTRACT(oFile.Name, This.cHomeDir,"",1,3)
		cNewFilePath = JUSTPATH(m.cNewName)
		IF NOT FILE(m.cNewName)
			IF NOT DIRECTORY(m.cNewFilePath)
				MD (m.cNewFilePath)
			ENDIF 
			COPY FILE (oFile.Name) TO (m.cNewName)
		ENDIF
		
		* Copy associated memo file.
		cFileExt = JUSTEXT(m.cNewName)
		IF INLIST(UPPER(m.cFileExt), "SCX", "VCX", "MNX", "FRX", "LBX")
			cNewExt = LEFT(m.cFileExt,2) + "T"
			COPY FILE (FORCEEXT(oFile.Name, m.cNewExt)) TO  (FORCEEXT(m.cNewName, m.cNewExt))
		ENDIF 
	ELSE
		* If the original file is not in the project
		* folder hierarchy, don't copy it. Just add the 
		* original to the new project.
		cNewName = oFile.Name
	ENDIF 
	
	* Now add it to the new project. Wrap in TRY-CATCH
	* in case it's already there.
	TRY
		oNewFile = oNewProject.Files.Add(m.cNewName)
		
		* Check whether it's the main file in the old project, and,\
		* if so, make it the main here.
		IF UPPER(oFile.Name) == UPPER(This.oProject.MainFile)
			oNewProject.SetMain(m.cNewName)
		ENDIF 
	CATCH
		cNoAdd = m.cNoAdd + CHR(13) + CHR(10) + m.cNewName
	ENDTRY
	
ENDFOR

* Modified 22-April-2011 by TEG
* Bring along icon, but fix path
LOCAL cIconPath, oIconNewPath
This.ReportToUser("Fixing project icon")
IF This.oProject.HomeDir $ This.oProject.Icon
	* First, make sure it was copied
	cNewName = m.cNewRoot + STREXTRACT(This.oProject.Icon, This.cHomeDir,"",1,3)
	IF NOT FILE(m.cNewname)
		cNewFilePath = JUSTPATH(m.cNewName)
		IF NOT DIRECTORY(m.cNewFilePath)
			MD (m.cNewFilePath)
		ENDIF 
		COPY FILE (This.oProject.Icon) TO (m.cNewName)
	ENDIF 
		
	cIconPath = FULLPATH(This.oProject.Icon, This.oProject.HomeDir)	
	* Now find the best relative reference for the new path
	cIconNewPath = SYS(2014, m.cIconPath, m.cNewRoot)	
	
	oNewProject.Icon = FORCEPATH(This.oProject.Icon, m.cIconNewPath)
ELSE
	oNewProject.Icon = This.oProject.Icon
ENDIF 

* Modified 22-April-2011 by TEG
* Look for missing graphics by scanning VCX and SCX
* files and identifying any Icon or Picture properties
* where the file is in the project hierarchy, but hasn't
* been copied. Copy them.
LOCAL cPictPath, cOldFilePath, cOldPictPath, cMissingPictures

This.ReportToUser("Ensuring all graphics are copied.") 
cMissingPictures = ""

FOR EACH oFile IN m.oNewProject.Files
	IF oFile.Type = "K" OR oFile.Type = "V"
		USE (oFile.Name) ALIAS __CXFile
		This.ReportToUser("Processing graphics in " + oFile.Name)
		cOldFilePath = ADDBS(This.cHomeDir) + STREXTRACT(oFile.Name, cNewRoot,"",1,3)
		SCAN FOR "Picture = " $ __CXFile.Properties OR "Icon = " $ __CXFile.Properties 
			IF "Picture = " $ __CXFile.Properties 
				cPictPath = STREXTRACT(__CXFile.Properties, "Picture = ", CHR(13) + CHR(10))
				* Get rid of relative pathing
				cOldPictPath = FULLPATH(m.cPictPath, m.cOldFilePath)
				IF UPPER(This.cHomeDir) $ UPPER(m.cOldPictPath)
					cNewName = m.cNewRoot + STREXTRACT(m.cOldPictPath, This.cHomeDir,"",1,3)
					IF NOT FILE(m.cNewName) 
						IF FILE(m.cOldPictPath) 
							IF NOT DIRECTORY(JUSTPATH(m.cNewName))
								MD (JUSTPATH(m.cNewName))
							ENDIF 
							COPY FILE (m.cOldPictPath) TO (m.cNewName)
						ELSE
							IF NOT (m.cOldPictPath + CHR(13) + CHR(10) $ m.cMissingPictures)
								cMissingPictures = cMissingPictures + m.cOldPictPath + CHR(13) + CHR(10)
							ENDIF 
						ENDIF 
					ENDIF 
				ENDIF
			ENDIF 

			IF "Icon = " $ __CXFile.Properties 
				cPictPath = STREXTRACT(__CXFile.Properties, "Icon = ", CHR(13) + CHR(10))
				cOldPictPath = FULLPATH(m.cPictPath, m.cOldFilePath)
				IF UPPER(This.cHomeDir) $ UPPER(m.cOldPictPath)
					cNewName = m.cNewRoot + STREXTRACT(m.cOldPictPath, This.cHomeDir,"",1,3)
					IF NOT FILE(m.cNewName) 
						IF FILE(m.cOldPictPath) 
							IF NOT DIRECTORY(JUSTPATH(m.cNewName))
								MD (JUSTPATH(m.cNewName))
							ENDIF 
							COPY FILE (m.cOldPictPath) TO (m.cNewName)
						ELSE
							IF NOT (m.cOldPictPath + CHR(13) + CHR(10) $ m.cMissingPictures)
								cMissingPictures = cMissingPictures + m.cOldPictPath + CHR(13) + CHR(10)
							ENDIF 
						ENDIF 
					ENDIF 
				ENDIF
			ENDIF 
		ENDSCAN
		
		USE IN SELECT("__CXFile")
	ENDIF
ENDFOR 

				

* Report to user
IF NOT EMPTY(m.cMissing)
	cMissing = "The following files in the project were not found: " + ;
	           m.cMissing
ENDIF 

IF NOT EMPTY(m.cNoAdd)
	cNoAdd = "The following files were not able to be added to the project: " + m.cNoAdd
ENDIF 

IF NOT EMPTY(m.cMissingPictures)
	cMissingPictures = "The following graphic files were not found: " + ;
	                   m.cMissingPictures
ENDIF 

DO CASE 
CASE EMPTY(m.cMissing) AND EMPTY(m.cNoAdd) 
	cMessage = "Project and all its files copied to " + m.cNewRoot
CASE EMPTY(m.cMissing)
	cMessage = "All project files were copied. " + m.cNoAdd
CASE EMPTY(m.cNoAdd)
	cMessage = m.cMissing
OTHERWISE
	* We have both kinds of message.
	cMessage = m.cMissing + CHR(13) + CHR(10) + m.cNoAdd
ENDCASE 

IF NOT EMPTY(m.cMissingPictures)
	cMessage = m.cMessage + m.cMissingPictures
ENDIF

MESSAGEBOX(m.cMessage)
	
* Close the new project
oNewProject.Close()

RETURN
ENDPROC 

****************************************************
* Progress routine for this class
****************************************************
PROCEDURE ReportToUser(cMessage)
* By pushing all progress reporting through this
* method, we can change the approach in one place.
* For now, just use WAIT WINDOWs.

IF EMPTY(m.cMessage)
	WAIT clear
ELSE 
	WAIT WINDOW m.cMessage NOWAIT 
ENDIF 

RETURN
ENDPROC

****************************************************
* Generic reporting routine for this class
****************************************************
PROCEDURE ReportOnProject(cAlias, cColumns, cReport)

IF VARTYPE(m.cAlias) <> "C" OR NOT USED(m.cAlias)
	ERROR 11
ENDIF 

LOCAL cReportFile, lCreatedReport, cFieldsClause
LOCAL aFieldList[1], nFields, nField, cFieldList
LOCAL aColumns[1], nColumns
LOCAL cFieldName, cNewFieldSpec

IF VARTYPE(m.cReport) <> "C" OR NOT FILE(m.cReport)

	cReportFile = FORCEPATH("ProjectReport",SYS(2023))
	lCreatedReport = .T.
	
	IF NOT EMPTY(m.cColumns)
		nColumns = ALINES(aColumns, m.cColumns, 1, ",")
	ELSE
		nColumns = 0
	ENDIF 
	
	* Handle memo fields, so that they get created wide enough
	nFields = AFIELDS(aFieldList, m.cAlias)
	cFieldList = ""
	
	FOR nField = 1 TO nFields
		cFieldName = aFieldList[m.nField, 1]
		IF m.nColumns = 0 OR ASCAN(aColumns, m.cFieldName, -1, -1, -1, 7) <> 0
			* It's in the list of columns to include
			IF aFieldList[m.nField,2] = "M" 
				* It's a memo field, so convert it to a long char for 
				* generating the report.
				cNewFieldSpec = "CAST(" + m.cFieldName + " AS C(250)) AS " + m.cFieldName
			ELSE
				* Just pass it along as is.
				cNewFieldSpec = m.cFieldName
			ENDIF 
			cFieldList = m.cFieldList + "," + m.cNewFieldSpec
		ENDIF 
	ENDFOR 
	
	cFieldList = SUBSTR(m.cFieldList, 2)
	
	* Now run the query
	SELECT &cFieldList ;
		FROM (m.cAlias) ;
		INTO CURSOR csrForQuickReport
	
	IF VARTYPE(m.cColumns) = "C" AND NOT EMPTY(m.cColumns)
		cFieldsClause = "FIELDS " + m.cColumns
	ELSE
		cFieldsClause = ""
	ENDIF 
		
	CREATE REPORT (m.cReportFile) FROM csrForQuickReport ;
		COLUMN &cFieldsClause.
		
	* Kill the special cursor
	USE IN SELECT("csrForQuickReport")
	
ELSE
	cReportFile = m.cReport
	lCreatedReport = .F.		
ENDIF 

SELECT (m.cAlias)
REPORT FORM (m.cReportFile) PREVIEW 

* Clean up
IF m.lCreatedReport
	ERASE (FORCEEXT(m.cReportFile, "FRX"))
	ERASE (FORCEEXT(m.cReportFile, "FRT"))
ENDIF

RETURN
ENDPROC 

****************************************************
* Methods below this comment are utilities that do
* not rely on properties of this class, and are
* thus candidates to be pulled out into 
* separate functions.
****************************************************
PROCEDURE GetValidAlias(cAlias, cDefault)
* Check whether the specified alias is valid.
* If so, return it. Otherwise, return the specified 
* default, if it's valid. Otherwise, return a
* randomly generated alias.

LOCAL cReturn

DO CASE 
CASE This.ValidAlias(m.cAlias)
	cReturn = m.cAlias
CASE This.ValidAlias(m.cDefault)
	cReturn = m.cDefault
OTHERWISE
	cReturn = SYS(2015)
ENDCASE 

RETURN m.cReturn
ENDPROC

PROCEDURE ValidAlias(cAlias)
* Check whether the parameter is a valid name
* for an alias. 

LOCAL lResult

lResult = .T.

DO CASE 

CASE VARTYPE(m.cAlias) <> "C"
	lResult = .F.
	
CASE LEN(m.cAlias) > 254
	lResult = .F.
	
CASE LEFT(m.cAlias,1) <> "_" AND NOT ISALPHA(LEFT(m.cAlias,1))
	lResult = .F.
	
CASE NOT EMPTY(CHRTRAN(LOWER(SUBSTR(m.cAlias, 2)),"_abcdefghijklmnopqrstuvwxyz1234567890",""))
	lResult = .F.
ENDCASE 

RETURN m.lResult
ENDPROC

ENDDEFINE
